/*
 *  Copyright: Collin Mulliner <collin[at]mulliner.org>
 *  Web: http://www.mulliner.org/nfc/
 *  
 *  License: GPLv2
 *  
 */

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Vector;

import javax.microedition.contactless.ContactlessException;
import javax.microedition.contactless.DiscoveryManager;
import javax.microedition.contactless.TargetListener;
import javax.microedition.contactless.TargetProperties;
import javax.microedition.contactless.TargetType;
import javax.microedition.contactless.ndef.NDEFTagConnection;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.StringItem;
import javax.microedition.lcdui.TextField;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;
import javax.microedition.rms.RecordEnumeration;
import javax.microedition.rms.RecordStore;

import com.nokia.nfc.nxp.mfstd.MFDataArea;
import com.nokia.nfc.nxp.mfstd.MFKey;
import com.nokia.nfc.nxp.mfstd.MFStandardConnection;
import com.nokia.nfc.nxp.simpletag.SimpleTagConnection;

public class BtNfcAdapterRaw extends MIDlet implements CommandListener, Runnable, TargetListener, DebugInterface
{
	private DiscoveryManager dm;
	private String uuid;
	private StreamConnectionNotifier btConNotifier;
	private StreamConnection clientCon;
	private DataInputStream clientInput;
	private DataOutputStream clientOutput;
	private boolean appRunning;
	
	static final char[] hexitem =  {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
	
	private Thread runner;
	
	private Command exitcmd;
	private Command clearcmd;
	private Command writeextcmd;
	private Command writeintcmd;
	
	private Form mainScreen;
	private TextField numReadTags;
	private TextField sizeWriteTag;
	//private TextField clientAddr;
	private TextField status;
	private TextField cardType;
	
	private int write;
	//private int writeInternal;
	private Vector readTags;
	private int numTags;
	
	//private NDEFMessage writeNdefMsg;
	private NFCTagCmd curCmd;
	private NFCTag writeTag;

	private RecordStore dataStore;
	
	private String dbg[];
	private int dbg_num;
	private int debug_level = 0;
	
	private MFKey rAKey = null;
	private byte[] rAKey_plain;
	private MFKey rBKey = null;
	private byte[] rBKey_plain;
	private MFKey wAKey = null;
	private byte[] wAKey_plain;
	private MFKey wBKey = null;
	private byte[] wBKey_plain;
	private int wkey_use; // 1 = A 2 = B
	private int rkey_use; // 1 = A 2 = B
	
	private static final String DBNAME = "TagsRaw";
	
	protected void destroyApp(boolean arg0) throws MIDletStateChangeException
	{
		appRunning = false;
		try {
			if (runner != null)
				runner.join();
		} catch (InterruptedException e) {}
	}
	
	public void debug(String d, int level)
	{
		StringBuffer ob = new StringBuffer();
		if (level == debug_level) {
			for (int i = 0; i < dbg_num - 1; i++) {
				dbg[i] = dbg[i+1];
			}
			dbg[dbg_num - 1] = d;
			for (int i = 0; i < dbg_num; i++) {
				if (dbg[i] != null) {
					ob.append(dbg[i] + "|");
				}
			}
			status.setString(ob.toString());
		}
	}
	
	private void save()
	{
		try {
			RecordStore.deleteRecordStore(DBNAME);
		} catch (Exception e) {}
		try {
			dataStore = RecordStore.openRecordStore(DBNAME, true);
			for (int i = 0; i < numTags; i++) {
				NFCTag t = (NFCTag) readTags.elementAt((numTags - 1) - i);
				dataStore.addRecord(t.toRMS(), 0, t.toRMS().length);
				debug("saved tag(s)", 0);
			}
			dataStore.closeRecordStore();
		} catch (Exception e) {}
	}

	public byte[] toHex(byte[] b)
	{
		//StringBuffer ob = new StringBuffer("");
		byte out[] = new byte[b.length * 2];
		
		for (int i = 0; i < b.length; i++) {
			int val = b[i] & 0xFF;
			out[i*2] = (byte) hexitem[val/16];
			out[i*2+1] = (byte) hexitem[val%16];
			//ob.append(hexitem[val/16]);
			//ob.append(hexitem[val%16]);
		}
		
		return out; //ob.toString();
	}
	
	public byte[] fromHex(String hex, int start)
	{
		int len = hex.length() - start;
		byte[] b = new byte[len/2];
		int bc = 0;
		
		//dbgout.debug(new String("len = " + String.valueOf(len) + " "), 3);
		
		for (int i = start; i < start + len; i+=2) {
			String cur = hex.substring(i, i+2);
			//dbgout.debug(cur, 3);
			b[bc] = (byte) Integer.parseInt(cur, 16);
			bc++;
		}
		
		return b;
	}
	
	protected void pauseApp() {}

	protected void startApp() throws MIDletStateChangeException
	{	
		exitcmd = new Command("Exit", Command.EXIT, 1);
		clearcmd = new Command("ClearTags", Command.ITEM, 1);
		writeextcmd = new Command("Write to External Tag", Command.ITEM, 1);
		writeintcmd = new Command("Write to Internal Tag", Command.ITEM, 1);
		mainScreen = new Form("RAW Bluetooth NFC Adapter");
		mainScreen.addCommand(exitcmd);
		mainScreen.addCommand(clearcmd);
		mainScreen.addCommand(writeextcmd);
		//mainScreen.addCommand(writeintcmd);
		
		mainScreen.append(new StringItem("Version:", "0.3.3"));
		mainScreen.append(new StringItem("Author:", "Collin Mulliner <collin@mulliner.org>"));
		mainScreen.append(new StringItem("Web:", "http://www.mulliner.org/nfc/"));
		mainScreen.append(new StringItem("License:", "GPLv2"));
		
		numReadTags = new TextField("No. Tags", "0", 3, TextField.UNEDITABLE);
		sizeWriteTag = new TextField("Write Size", "0", 5, TextField.UNEDITABLE);
		//clientAddr = new TextField("Client", "", 20, TextField.UNEDITABLE);
		status = new TextField("Status", "started", 1024, TextField.UNEDITABLE);
		cardType = new TextField("Card Type:", "", 50, TextField.UNEDITABLE);
		
		//mainScreen.append(clientAddr);
		mainScreen.append(numReadTags);
		mainScreen.append(sizeWriteTag);
		mainScreen.append(cardType);
		mainScreen.append(status);
		
		Display.getDisplay(this).setCurrent(mainScreen);
	
		mainScreen.setCommandListener(this);
		registerTargetListner();
		
		uuid = new String("AABBCCDDAABBCCDDAABBCCDDAABBCCDD");
		write = 0;
		//writeInternal = 0;
		appRunning = true;
		readTags = new Vector();
		numTags = 0;
		
		dbg_num = 5;
		dbg = new String[dbg_num];
		
		try {
			dataStore = RecordStore.openRecordStore(DBNAME, true);
			RecordEnumeration re = dataStore.enumerateRecords(null, null, false);
			while (re.hasNextElement()) {
				byte[] b = re.nextRecord();
				NFCTag t = new NFCTag();
				t.fromRMS(b);
				readTags.addElement(t);
				numTags++;
			}
			dataStore.closeRecordStore();
		} catch (Exception e) {
			//System.out.println("loading err: " + e.getMessage());
		}
		numReadTags.setString(String.valueOf(numTags));
		
		runner = new Thread(this);
		runner.start();
	}

	public void commandAction(Command cmd, Displayable disp)
	{
		if (cmd == exitcmd) {
			save();
			notifyDestroyed();
		}
		else if (cmd == clearcmd) {
			readTags.removeAllElements();
			numTags = 0;
			numReadTags.setString(String.valueOf(numTags));
			debug("tag cache cleared", 0);
		}
		else if (cmd == writeextcmd) {
			if (numTags > 0) {
				writeTag = (NFCTag) readTags.elementAt(0);
				sizeWriteTag.setString(String.valueOf(writeTag.getHex().length));
				//writeNdefMsg = writeTag.message;
				write = 1;
				readTags.removeElementAt(0);
				numTags--;
				numReadTags.setString(String.valueOf(numTags));
			}
		}
		else if (cmd == writeintcmd) {
			// doesn't work yet
			writeTag = (NFCTag) readTags.elementAt(0);
			//writeNdefMsg = writeTag.message;
			readTags.removeElementAt(0);
			numTags--;
			numReadTags.setString(String.valueOf(numTags));
			debug("write internal tag ", 0);
			try {
				String uri = System.getProperty("internal.mf.url");
				debug("uri=" + uri + " ", 0);
				NDEFTagConnection conn = (NDEFTagConnection) Connector.open(uri);
				if (conn != null) {
					//conn.writeNDEF(writeNdefMsg);
					conn.close();
				}
				else
					debug("internal connection failed ", 0);
			} catch (Exception e) {
				debug("int write err: " + e.getMessage() + " ", 0);
			}
			//writeNdefMsg = null;
		}
	}
	
	private boolean waitForClient()
	{
		try {
			clientCon = btConNotifier.acceptAndOpen();
			clientInput = new DataInputStream(clientCon.openInputStream());
			clientOutput = new DataOutputStream(clientCon.openOutputStream());
		}
		catch (IOException e) {
			return false;
		}
		debug("client connected ", 0);
		return true;
	}
	
	private boolean sendNull()
	{
		String outp = new String("0\n");
		try {
			clientOutput.write(outp.getBytes());
			clientOutput.flush();
			return true;
		} catch (IOException e) {
			return false;
		}
	}
	
	public void run()
	{
		String server_url = new String("btspp://localhost:" + uuid + ";name=BtNfcAdapter");
		CommandInterface ci = null;
	
		try {
			btConNotifier = (StreamConnectionNotifier) Connector.open(server_url);
		} catch (IOException e) {}
		
		boolean closeCon = false;
		while (appRunning) {
			if (clientCon == null) {
				if (waitForClient())
					ci = new CommandInterface(clientInput, this);
				else
					clientCon = null;
			}
			else {
				debug("wait for cmd ", 0);
				curCmd = ci.getCommand();
				if (curCmd == null)
					closeCon = true;
				else {
					// read
					if (curCmd.getType() == 1) {
						if (readTags.isEmpty()) {
							debug("got Read Command, but no Tags read so far ", 0);
							if (!sendNull())
								closeCon = true;
						}
						else {
							debug("got Read Command ", 0);
							NFCTag tag = (NFCTag) readTags.elementAt(0);
							if (tag != null) {
								readTags.removeElementAt(0);
								numTags--;
								try {
									String outp = tag.toReadTagCmd();
									clientOutput.write(outp.getBytes());
									clientOutput.write(tag.getHex());
									clientOutput.write(new String("\n").getBytes());
									clientOutput.flush();
								} catch (IOException e) {
									debug(" error sending tag data to client: " + e.toString(), 0);
									closeCon = true;
								}
							}
							else {
								debug("got Read Command, tag is null ", 0);
								if (!sendNull())
									closeCon = true;
							}
							numReadTags.setString(String.valueOf(numTags));
						}
					}
					// write
					else if (curCmd.getType() == 2) {
						debug("got Write Command ", 0);
						writeTag = new NFCTag(curCmd.getData(), this);
						sizeWriteTag.setString(String.valueOf(curCmd.getData().length()));
						//writeNdefMsg = writeTag.message;
						write = 1;
						//if (writeNdefMsg == null) {
						if (writeTag == null) {
							debug("writeNdefMsg = null ", 0);
							write = 0;
							sizeWriteTag.setString("0");
						}
					}
					else if (curCmd.getType() == 3) { // WriteMFTag
						debug("got MF Write Command ", 0);
						writeTag = new NFCTag(curCmd.getData(), this);
						sizeWriteTag.setString(String.valueOf(curCmd.getData().length()));
						//writeNdefMsg = writeTag.message;
						write = 2;
						//if (writeNdefMsg == null) {
						if (writeTag == null) {
							debug("writeNdefMsg = null ", 0);
							write = 0;
							sizeWriteTag.setString("0");
						}
					}
					else if (curCmd.getType() == 4) { // set Read AKey
						rAKey_plain = fromHex(curCmd.key, 0);
						rAKey = new MFKey.KeyA(rAKey_plain);
						rkey_use = 1;
					}
					else if (curCmd.getType() == 5) { // set Read BKey
						rBKey_plain = fromHex(curCmd.key, 0);
						rBKey = new MFKey.KeyB(rBKey_plain);
						rkey_use = 2;
					}
					else if (curCmd.getType() == 6) { // set Write AKey
						wAKey_plain = fromHex(curCmd.key, 0);
						wAKey = new MFKey.KeyA(wAKey_plain);
						wkey_use = 1;
					}
					else if (curCmd.getType() == 7) { // set Write BKey
						wBKey_plain = fromHex(curCmd.key, 0);
						wBKey = new MFKey.KeyA(wBKey_plain);
						wkey_use = 2;
					}
				}
			}
			if (closeCon) {
				try {
					debug("client disconnected ", 0);
					clientInput.close();
					clientOutput.close();
					clientCon.close();
				} catch (IOException e) {}
				closeCon = false;
				clientCon = null;
			}
			// cleanup memory
			System.gc();
		}
	}
	
	private void targetDetected_read(TargetProperties prop, Class con, int ct)
	{
		NFCTag curTag = new NFCTag();
		
		curTag.mapping = prop.getMapping();
		curTag.uid = prop.getUid();
		curTag.url = prop.getUrl();
		curTag.targetTypes = prop.getTargetTypes();
		curTag.dateTime = new Date();
		
		Class[] tempcn = prop.getConnectionNames();
		curTag.connNames = new String[tempcn.length];
		for (int i = 0; i < tempcn.length; i++) {
			curTag.connNames[i] = tempcn[i].toString();
		}
		
		cardType.setString(curTag.getCardType(this));
		
		SimpleTagConnection sconn = null;
		MFStandardConnection mconnc = null;
		
		try {
			if (ct == 1) {
				sconn = (SimpleTagConnection) Connector.open(prop.getUrl(con));
			
				curTag.tag_size = sconn.getUserDataSize();
				curTag.data = sconn.readLogical(0, curTag.tag_size);
			}
			else if (ct == 2) {
				debug("Mifare Tag ", 0);
				MFDataArea mconn = (MFDataArea) Connector.open(prop.getUrl(con));
				mconnc = (MFStandardConnection) mconn;
				curTag.tag_size = mconn.size();
				curTag.data = new byte[curTag.tag_size];
				mconn.read(rkey_use == 1 ? rAKey : rBKey, curTag.data, 0, 0, curTag.tag_size);
			}
		}
		catch (ContactlessException ce) {
			debug(ce.getMessage() + " ", 0);
			curTag.state = -1;
		}
		catch (Exception e) {
			debug("read err: " + e.getMessage(), 0);
			curTag.state = -1;
		}		
		
		try {
			if (sconn != null)
				sconn.close();
			if (mconnc != null)
				mconnc.close();
		} catch (Exception e) {}
		
		if (curTag.state != -1) {
			readTags.addElement(curTag);
			numTags++;
			debug("tag read ", 0);
			numReadTags.setString(String.valueOf(numTags));
		}
	}
	
	private void targetDetected_write(TargetProperties prop, Class con, int ct)
	{
		//NDEFTagConnection conn = null;
		SimpleTagConnection sconn = null;
		MFDataArea mconn = null;
		
		if (writeTag == null)
			return;
		
		// disable write - not implemented yet
		if (true) return;
		
		try {
			//conn = (NDEFTagConnection) Connector.open(prop[0].getUrl());
			//conn.writeNDEF(writeNdefMsg);
			
			if (ct == 1) {
				debug("Write Simple Tag ", 0);
				sconn = (SimpleTagConnection) Connector.open(prop.getUrl(con));
				sconn.writeLogical(0, writeTag.data);
			}
			else if (ct == 2) {
				debug("Write Mifare Tag ", 0);
				mconn = (MFDataArea) Connector.open(prop.getUrl(con));
				
				int dl = writeTag.data.length;
				int block = 4; // we always skip the first sector
				int sp = 0;
				
				while (dl > 0) {
					if (dl >= 16)
						mconn.write(wkey_use == 1 ? wAKey : wBKey, writeTag.data, (sp * 16), 16, (block * 16));
					else
						mconn.write(wkey_use == 1 ? wAKey : wBKey, writeTag.data, (sp * 16), dl, (block * 16));
					dl -= 16;
					block++;
					// skip sector trailers
					if (block < 128 && block % 4 == 3)
						block++;
					else if (block >= 128 && block % 8 == 7)
						block++;
					sp++;
				}
			}
        }
		catch (ContactlessException e) {
			debug("write err: " + e.toString(), 0);
        }
		catch (IOException e) {
			debug("write err: " + e.toString(), 0);
        }
		finally {
			if (sconn != null) {
				try {
					sconn.close();
				} catch (IOException e) {
					debug("write err: " + e.toString(), 0);
                }
            }
		}
		//writeNdefMsg = null;
		writeTag = null;
		debug("tag written ", 0);
		sizeWriteTag.setString("0");
	}
	
	public void targetDetected(TargetProperties[] prop)
	{
		Class cn[] = prop[0].getConnectionNames();
		Class con = null;
		int ct = 0;
		
		for (int i = 0; i < cn.length; i++) {
			//debug(cn[i].toString(), 0);
			if (cn[i].toString().indexOf("SimpleConnection") > 0) {
				con = cn[i];
				ct = 1;
				break;
			}
			if (cn[i].toString().indexOf("MFStandardConnection") > 0) {
				con = cn[i];
				ct = 2;
				break;
			}
		}
		if (con == null) {
			return;
		}
		
		if (write == 0) {
			targetDetected_read(prop[0], con, ct);
		}
		else {
			targetDetected_write(prop[0], con, ct);
			write = 0;
		}	
	}
	
	private void registerTargetListner()
	{
		dm = DiscoveryManager.getInstance();
		try {
			dm.addTargetListener(this, TargetType.NDEF_TAG);
			dm.addTargetListener(this, TargetType.ISO14443_CARD);
			dm.addTargetListener(this, TargetType.RFID_TAG);
		}
		catch (IllegalStateException e) {}
		catch (ContactlessException e) {}
	}
}